# 49 "src/kernel_internals/parsing/clexer.mll"
 
open Cparser
module H = Hashtbl
module E = Errorloc

let currentLoc () = Errorloc.currentLoc ()

let one_line_ghost = ref false
let is_oneline_ghost () = !one_line_ghost
let enter_oneline_ghost () = one_line_ghost := true
let exit_oneline_ghost () = one_line_ghost := false

let ghost_code = ref false
let is_ghost_code () = !ghost_code
let enter_ghost_code () = ghost_code := true
let exit_ghost_code () = ghost_code := false

let addComment c = Cabshelper.Comments.add (currentLoc()) c

(* Some debugging support for line numbers *)
let dbgToken (t: token) =
  if false then begin
    let dprintf fmt = Kernel.debug fmt in
    (match t with
         IDENT n -> dprintf "IDENT(%s)\n" n
       | LBRACE l -> dprintf "LBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | RBRACE l -> dprintf "RBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | IF l -> dprintf "IF(%d)\n" (fst l).Lexing.pos_lnum
       | SWITCH l -> dprintf "SWITCH(%d)\n" (fst l).Lexing.pos_lnum
       | RETURN l -> dprintf "RETURN(%d)\n" (fst l).Lexing.pos_lnum
       | _ -> ()) ;
    t
  end else
    t

(*
** Keyword hashtable
*)
let lexicon = H.create 211
let init_lexicon _ =
  H.clear lexicon;
  Logic_env.reset_typenames ();
  Logic_env.builtin_types_as_typenames ();
  List.iter
    (fun (key, builder) -> H.add lexicon key builder)
    [ ("auto", fun loc -> AUTO loc);
      ("const", fun loc -> CONST loc);
      ("__const", fun loc -> CONST loc);
      ("__const__", fun loc -> CONST loc);
      ("static", fun loc -> STATIC loc);
      ("extern", fun loc -> EXTERN loc);
      ("long", fun loc -> LONG loc);
      ("short", fun loc -> SHORT loc);
      ("register", fun loc -> REGISTER loc);
      ("signed", fun loc -> SIGNED loc);
      ("__signed", fun loc -> SIGNED loc);
      ("unsigned", fun loc -> UNSIGNED loc);
      ("volatile", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);
      (* WW: see /usr/include/sys/cdefs.h for why __signed and __volatile
       * are accepted GCC-isms *)
      ("char", fun loc -> CHAR loc);
      ("_Bool", fun loc -> BOOL loc);
      ("int", fun loc -> INT loc);
      ("float", fun loc -> FLOAT loc);
      ("double", fun loc -> DOUBLE loc);
      ("void", fun loc -> VOID loc);
      ("enum", fun loc -> ENUM loc);
      ("struct", fun loc -> STRUCT loc);
      ("typedef", fun loc -> TYPEDEF loc);
      ("union", fun loc -> UNION loc);
      ("break", fun loc -> BREAK loc);
      ("continue", fun loc -> CONTINUE loc);
      ("goto", fun loc -> GOTO loc);
      ("return", fun loc -> dbgToken (RETURN loc));
      ("switch", fun loc -> dbgToken (SWITCH loc));
      ("case", fun loc -> CASE loc);
      ("default", fun loc -> DEFAULT loc);
      ("while", fun loc -> WHILE loc);
      ("do", fun loc -> DO loc);
      ("for", fun loc -> FOR loc);
      ("if", fun loc -> dbgToken (IF loc));
      ("else", fun _ -> ELSE);
      (*** Implementation specific keywords ***)
      ("__signed__", fun loc -> SIGNED loc);
      ("__inline__", fun loc -> INLINE loc);
      ("inline", fun loc -> INLINE loc);
      ("__inline", fun loc -> INLINE loc);
      ("_inline", fun loc ->
                      if !Cprint.msvcMode then
                        INLINE loc
                      else
                        IDENT ("_inline"));
      ("__attribute__", fun loc -> ATTRIBUTE loc);
      ("__attribute", fun loc -> ATTRIBUTE loc);
      ("__blockattribute__", fun _ -> BLOCKATTRIBUTE);
      ("__blockattribute", fun _ -> BLOCKATTRIBUTE);
      ("__asm__", fun loc -> ASM loc);
      ("asm", fun loc -> ASM loc);
      ("__typeof__", fun loc -> TYPEOF loc);
      ("__typeof", fun loc -> TYPEOF loc);
      ("typeof", fun loc -> TYPEOF loc);
      ("__alignof", fun loc -> ALIGNOF loc);
      ("__alignof__", fun loc -> ALIGNOF loc);
      ("__volatile__", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);

      ("__FUNCTION__", fun loc -> FUNCTION__ loc);
      ("__func__", fun loc -> FUNCTION__ loc); (* ISO 6.4.2.2 *)
      ("__PRETTY_FUNCTION__", fun loc -> PRETTY_FUNCTION__ loc);
      ("__label__", fun _ -> LABEL__);
      (*** weimer: GCC arcana ***)
      ("__restrict", fun loc -> RESTRICT loc);
      ("restrict", fun loc -> RESTRICT loc);
(*      ("__extension__", EXTENSION); *)
      (**** MS VC ***)
      ("__int64", fun _ -> INT64 (currentLoc ()));
      ("__int32", fun loc -> INT loc);
      ("_cdecl",  fun _ -> MSATTR ("_cdecl", currentLoc ()));
      ("__cdecl", fun _ -> MSATTR ("__cdecl", currentLoc ()));
      ("_stdcall", fun _ -> MSATTR ("_stdcall", currentLoc ()));
      ("__stdcall", fun _ -> MSATTR ("__stdcall", currentLoc ()));
      ("_fastcall", fun _ -> MSATTR ("_fastcall", currentLoc ()));
      ("__fastcall", fun _ -> MSATTR ("__fastcall", currentLoc ()));
      ("__w64", fun _ -> MSATTR("__w64", currentLoc ()));
      ("__declspec", fun loc -> DECLSPEC loc);
      ("__forceinline", fun loc -> INLINE loc); (* !! we turn forceinline
                                                 * into inline *)
      ("__try", fun loc -> TRY loc);
      ("__except", fun loc -> EXCEPT loc);
      ("__finally", fun loc -> FINALLY loc);
      (* weimer: some files produced by 'GCC -E' expect this type to be
       * defined *)
      ("__builtin_va_list",
       fun _ -> NAMED_TYPE "__builtin_va_list");
      ("__builtin_va_arg", fun loc -> BUILTIN_VA_ARG loc);
      ("__builtin_types_compatible_p", fun loc -> BUILTIN_TYPES_COMPAT loc);
      ("__builtin_offsetof", fun loc -> BUILTIN_OFFSETOF loc);
      (* On some versions of GCC __thread is a regular identifier *)
      ("__thread",
       (fun loc ->
          if Cil.theMachine.Cil.theMachine.Cil_types.__thread_is_keyword then
            THREAD loc
          else
            IDENT "__thread"));
    ]


let is_c_keyword s = Hashtbl.mem lexicon s

(* Mark an identifier as a type name. The old mapping is preserved and will
 * be reinstated when we exit this context *)
let add_type name =
   (* ignore (print_string ("adding type name " ^ name ^ "\n"));  *)
  H.add lexicon name (fun _ -> NAMED_TYPE name);
  Logic_env.add_typename name

let context : string list list ref = ref [ [] ]

let push_context _ = context := []::!context

let pop_context _ =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
		(context := sub;
		List.iter (fun name ->
                            (* Format.eprintf
                              "removing lexicon for %s@." name; *)
                             H.remove lexicon name;
                             Logic_env.remove_typename name
                          ) con)

(* Mark an identifier as a variable name. The old mapping is preserved and
 * will be reinstated when we exit this context  *)
let add_identifier name =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
      (context := (name::con)::sub;
       (*Format.eprintf "adding IDENT for %s@." name;*)
       H.add lexicon name (fun _ -> dbgToken (IDENT name));
       Logic_env.hide_typename name
      )


(*
** Useful primitives
*)
let scan_ident id =
  let here = currentLoc () in
  try (H.find lexicon id) here
  (* default to variable name, as opposed to type *)
  with Not_found -> dbgToken (IDENT id)


(*
** Buffer processor
*)


let init ~(filename: string) : Lexing.lexbuf =
  init_lexicon ();
  (* Inititialize the pointer in Errormsg *)
  Lexerhack.add_type := add_type;
  Lexerhack.push_context := push_context;
  Lexerhack.pop_context := pop_context;
  Lexerhack.add_identifier := add_identifier;
  E.startParsing filename


let finish () =
  E.finishParsing ();
  Logic_env.reset_typenames ()


(*** escape character management ***)
let scan_escape (char: char) : int64 =
  let result = match char with
    'n' -> '\n'
  | 'r' -> '\r'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'f' -> '\012'  (* ASCII code 12 *)
  | 'v' -> '\011'  (* ASCII code 11 *)
  | 'a' -> '\007'  (* ASCII code 7 *)
  | 'e' | 'E' -> '\027'  (* ASCII code 27. This is a GCC extension *)
  | '\'' -> '\''
  | '"'-> '"'     (* '"' *)
  | '?' -> '?'
  | '(' when not !Cprint.msvcMode -> '('
  | '{' when not !Cprint.msvcMode -> '{'
  | '[' when not !Cprint.msvcMode -> '['
  | '%' when not !Cprint.msvcMode -> '%'
  | '\\' -> '\\'
  | other ->
    E.parse_error "Unrecognized escape sequence: \\%c" other
  in
  Int64.of_int (Char.code result)

let scan_hex_escape str =
  let radix = Int64.of_int 16 in
  let the_value = ref Int64.zero in
  (* start at character 2 to skip the \x *)
  for i = 2 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 16 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let scan_oct_escape str =
  let radix = Int64.of_int 8 in
  let the_value = ref Int64.zero in
  (* start at character 1 to skip the \x *)
  for i = 1 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 8 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let lex_hex_escape remainder lexbuf =
  let prefix = scan_hex_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_oct_escape remainder lexbuf =
  let prefix = scan_oct_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_simple_escape remainder lexbuf =
  let lexchar = Lexing.lexeme_char lexbuf 1 in
  let prefix = scan_escape lexchar in
  prefix :: remainder lexbuf

let lex_unescaped remainder lexbuf =
  let prefix = Int64.of_int (Char.code (Lexing.lexeme_char lexbuf 0)) in
  prefix :: remainder lexbuf

let lex_comment remainder buffer lexbuf =
  let ch = Lexing.lexeme_char lexbuf 0 in
  if ch = '\n' then E.newline() ;
  (match buffer with None -> () | Some b -> Buffer.add_char b ch) ;
  remainder buffer lexbuf

let do_lex_comment ?(first_string="") remainder lexbuf =
  let buffer =
    if Kernel.PrintComments.get () then begin
      let b = Buffer.create 80 in
      Buffer.add_string b first_string;
      Some b
    end else None
  in remainder buffer lexbuf ;
  match buffer with
    | Some b -> addComment (Buffer.contents b)
    | None -> ()


(* ISO standard locale-specific function to convert a wide character
 * into a sequence of normal characters. Here we work on strings.
 * We convert L"Hi" to "H\000i\000"
  matth: this seems unused.
let wbtowc wstr =
  let len = String.length wstr in
  let dest = String.make (len * 2) '\000' in
  for i = 0 to len-1 do
    dest.[i*2] <- wstr.[i] ;
  done ;
  dest
*)

(* This function converst the "Hi" in L"Hi" to { L'H', L'i', L'\0' }
  matth: this seems unused.
let wstr_to_warray wstr =
  let len = String.length wstr in
  let res = ref "{ " in
  for i = 0 to len-1 do
    res := !res ^ (Printf.sprintf "L'%c', " wstr.[i])
  done ;
  res := !res ^ "}" ;
  !res
*)

(* Pragmas get explicit end-of-line tokens.
 * Elsewhere they are silently discarded as whitespace. *)
let pragmaLine = ref false

let annot_char = ref '@'

let () =
  Kernel.ReadAnnot.add_set_hook
    (fun _ x ->
      (* prevent the C lexer interpretation of comments *)
      annot_char := if x then '@' else '\000');
  Kernel.CustomAnnot.add_set_hook (fun _ s -> annot_char:=s.[0])

let annot_start_pos = ref Cabshelper.cabslu
let buf = Buffer.create 1024

let save_current_pos () =
  annot_start_pos := currentLoc ()

let make_annot ~one_line lexbuf s =
  let start = snd !annot_start_pos in
  let stop, token = Logic_lexer.annot (start, s) in
  lexbuf.Lexing.lex_curr_p <- stop; 
  (* The filename has already been normalized, so we must reuse it "as is". *)
  E.setCurrentFile ~normalize:false stop.Lexing.pos_fname;
  E.setCurrentLine stop.Lexing.pos_lnum;
  if one_line then E.newline ();
  match token with
    | Logic_ptree.Adecl d -> DECL d
    | Logic_ptree.Aspec -> SPEC (start,s)
        (* At this point, we only have identified a function spec. Complete
           parsing of the annotation will only occur in the cparser.mly rule.
         *)
    | Logic_ptree.Acode_annot (loc,a) -> CODE_ANNOT (a, loc)
    | Logic_ptree.Aloop_annot (loc,a) -> LOOP_ANNOT (a,loc)
    | Logic_ptree.Aattribute_annot (loc,a) -> ATTRIBUTE_ANNOT (a, loc)
    | Logic_ptree.Acustom(loc,id, a) -> CUSTOM_ANNOT(a, id, loc)


# 365 "src/kernel_internals/parsing/clexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\179\255\180\255\091\000\192\000\187\255\188\255\189\255\
    \190\255\193\255\194\255\197\255\198\255\199\255\002\000\201\255\
    \035\000\067\000\093\000\070\000\094\000\091\000\072\000\169\000\
    \173\000\074\000\013\001\025\001\063\001\243\255\136\001\245\255\
    \223\001\076\000\248\255\001\000\250\255\074\001\175\000\231\255\
    \173\000\092\001\254\255\059\000\255\255\252\255\095\000\253\255\
    \249\255\005\000\195\255\230\255\247\255\054\002\141\002\228\002\
    \059\003\146\003\233\003\064\004\151\004\238\004\069\005\156\005\
    \243\005\074\006\161\006\248\006\079\007\166\007\253\007\084\008\
    \171\008\002\009\089\009\176\009\007\010\094\010\181\010\012\011\
    \242\255\244\255\101\000\089\011\117\011\127\011\082\000\052\001\
    \062\001\139\011\161\011\131\000\156\011\158\011\173\011\183\000\
    \077\001\076\001\058\001\167\000\098\001\153\011\175\011\199\011\
    \169\000\241\255\004\012\014\012\043\012\053\012\212\001\082\012\
    \130\002\092\012\217\002\153\012\191\012\169\000\178\000\008\001\
    \163\011\121\012\005\013\229\012\038\013\048\003\071\013\135\003\
    \239\255\168\000\148\000\117\000\174\000\078\012\173\000\240\000\
    \020\001\165\011\240\255\018\001\156\000\125\000\183\000\192\000\
    \235\255\221\255\122\000\145\000\135\000\003\000\236\255\215\255\
    \234\255\213\255\214\255\233\255\232\255\205\255\229\255\206\255\
    \228\255\227\255\192\255\196\255\220\255\194\000\226\255\219\255\
    \227\000\225\255\222\255\191\255\115\013\202\013\033\014\120\014\
    \207\014\025\001\255\255\094\001\253\255\254\255\038\001\255\255\
    \097\001\254\255\255\255\065\015\249\255\250\255\046\001\252\255\
    \253\255\086\002\255\255\251\255\070\015\249\255\232\000\242\000\
    \080\015\009\004\255\255\041\001\122\001\064\001\175\011\253\255\
    \065\001\021\001\247\000\072\001\031\001\042\001\252\255\057\001\
    \052\001\049\001\068\001\096\004\129\015\058\001\066\001\109\001\
    \094\001\081\001\082\001\088\001\127\001\129\001\165\001\184\001\
    \173\001\168\001\173\001\179\001\176\001\184\001\251\255\207\001\
    \248\001\215\001\207\001\203\001\248\001\233\001\247\001\245\001\
    \247\001\253\001\009\002\004\002\006\002\008\002\044\002\051\002\
    \032\002\047\002\051\002\072\002\094\002\089\002\091\002\102\002\
    \102\002\102\002\100\002\122\002\132\002\139\002\134\002\117\002\
    \159\002\176\002\190\002\179\002\188\002\207\002\196\002\195\002\
    \211\002\215\002\218\002\212\002\213\002\008\003\020\003\010\003\
    \024\003\006\003\020\003\019\003\018\003\035\003\035\003\051\003\
    \039\003\056\003\096\003\088\003\100\003\095\003\097\003\124\013\
    \251\255\189\011\183\004\255\255\252\255\105\015\134\015\136\015\
    \253\255\198\003\253\255\254\255\255\255\026\004\254\255\255\255\
    \151\010\250\255\251\255\177\012\255\255\252\255\058\005\153\015\
    \208\015\145\005\253\255\147\010\250\255\251\255\231\015\255\255\
    \252\255\232\005\246\015\058\016\063\006\253\255\037\013\253\255\
    \254\255\193\006\155\003\254\255\255\255\011\011\253\255\114\003\
    \255\255\113\003\122\003\129\003\254\255\007\004\253\255\254\255\
    \135\003\129\003\127\003\154\003\255\255\027\004\252\255\253\255\
    \254\255\224\003\255\255\094\004\254\255\171\003\168\003\179\003\
    \180\003\255\255\114\004\254\255\255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\074\000\074\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\055\000\255\255\
    \037\000\039\000\038\000\053\000\051\000\052\000\045\000\044\000\
    \043\000\048\000\069\000\018\000\018\000\255\255\074\000\255\255\
    \074\000\047\000\255\255\076\000\255\255\004\000\046\000\255\255\
    \002\000\000\000\255\255\001\000\255\255\255\255\003\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\074\000\074\000\074\000\
    \074\000\074\000\074\000\009\000\074\000\074\000\074\000\074\000\
    \071\000\074\000\074\000\074\000\074\000\072\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\073\000\
    \255\255\255\255\255\255\255\255\255\255\014\000\255\255\018\000\
    \018\000\018\000\017\000\255\255\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\255\255\018\000\018\000\018\000\018\000\
    \255\255\255\255\255\255\014\000\255\255\014\000\255\255\014\000\
    \255\255\014\000\255\255\255\255\016\000\255\255\016\000\016\000\
    \016\000\255\255\255\255\255\255\014\000\255\255\014\000\255\255\
    \255\255\016\000\016\000\016\000\255\255\015\000\255\255\015\000\
    \015\000\015\000\255\255\015\000\015\000\015\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\031\000\255\255\255\255\
    \032\000\255\255\255\255\255\255\074\000\074\000\074\000\074\000\
    \070\000\001\000\255\255\255\255\255\255\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\006\000\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\006\000\006\000\
    \002\000\001\000\255\255\255\255\002\000\002\000\002\000\255\255\
    \002\000\002\000\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\004\000\001\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\255\255\255\255\002\000\003\000\
    \001\000\002\000\255\255\255\255\255\255\255\255\005\000\255\255\
    \255\255\002\000\003\000\001\000\002\000\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\002\000\
    \255\255\255\255\255\255\255\255\255\255\003\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\002\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\000\000\
    \255\255\255\255\000\000\255\255\000\000\255\255\255\255\000\000\
    \045\000\042\000\000\000\255\255\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \255\255\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\180\000\000\000\000\000\255\255\000\000\
    \185\000\000\000\000\000\188\000\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000\000\000\197\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\048\001\
    \000\000\053\001\255\255\000\000\000\000\053\001\053\001\053\001\
    \000\000\058\001\000\000\000\000\000\000\062\001\000\000\000\000\
    \065\001\000\000\000\000\069\001\000\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\076\001\000\000\000\000\080\001\000\000\
    \000\000\255\255\255\255\255\255\255\255\000\000\087\001\000\000\
    \000\000\255\255\091\001\000\000\000\000\094\001\000\000\255\255\
    \000\000\255\255\255\255\255\255\000\000\255\255\000\000\000\000\
    \255\255\255\255\255\255\255\255\000\000\110\001\000\000\000\000\
    \000\000\255\255\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\000\000\123\001\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\037\000\036\000\048\000\037\000\037\000\049\000\048\000\
    \000\000\000\000\049\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\037\000\000\000\000\000\000\000\000\000\000\000\
    \037\000\025\000\029\000\034\000\150\000\033\000\020\000\031\000\
    \008\000\007\000\022\000\024\000\005\000\023\000\026\000\038\000\
    \028\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\014\000\006\000\018\000\016\000\017\000\015\000\
    \171\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\030\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\010\000\035\000\009\000\019\000\032\000\
    \170\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\004\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\012\000\021\000\011\000\013\000\003\000\
    \167\000\168\000\163\000\161\000\159\000\156\000\052\000\145\000\
    \104\000\051\000\050\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\133\000\133\000\162\000\
    \158\000\165\000\164\000\160\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\044\000\255\255\
    \044\000\099\000\003\000\146\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\154\000\157\000\
    \151\000\041\000\047\000\095\000\047\000\100\000\040\000\132\000\
    \128\000\128\000\128\000\142\000\003\000\085\000\155\000\153\000\
    \138\000\138\000\152\000\138\000\039\000\046\000\144\000\147\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\148\000\149\000\084\000\128\000\130\000\166\000\
    \002\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\117\000\084\000\128\000\131\000\003\000\
    \169\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\172\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\143\000\140\000\107\000\107\000\107\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\085\000\
    \178\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\037\000\129\000\183\000\037\000\037\000\
    \195\000\134\000\215\000\212\000\141\000\128\000\084\000\211\000\
    \139\000\207\000\085\000\207\000\037\000\088\000\255\255\138\000\
    \085\000\138\000\037\000\186\000\085\000\085\000\087\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\027\000\
    \027\000\084\000\085\000\085\000\207\000\128\000\084\000\084\000\
    \102\000\082\000\086\000\084\000\084\000\089\000\255\255\138\000\
    \182\000\138\000\101\000\088\000\208\000\213\000\087\000\214\000\
    \085\000\084\000\084\000\100\000\087\000\207\000\207\000\083\000\
    \095\000\084\000\216\000\217\000\043\000\086\000\218\000\084\000\
    \103\000\082\000\095\000\084\000\084\000\219\000\095\000\084\000\
    \086\000\041\001\080\000\089\000\003\000\255\255\028\001\081\000\
    \021\001\084\000\084\000\100\000\087\000\207\000\207\000\083\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\095\000\007\001\254\000\253\000\209\000\084\000\
    \249\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\203\000\244\000\241\000\210\000\003\000\
    \240\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\239\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\233\000\234\000\235\000\236\000\232\000\237\000\238\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\054\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\238\000\238\000\242\000\243\000\053\000\238\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\245\000\255\255\246\000\181\000\193\000\
    \247\000\186\000\193\000\193\000\248\000\238\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \193\000\250\000\251\000\252\000\238\000\238\000\193\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\255\000\000\001\001\001\002\001\003\000\003\001\062\000\
    \003\000\003\000\003\000\060\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\061\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\111\000\111\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\004\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\005\001\
    \006\001\238\000\008\001\009\001\010\001\011\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \012\001\013\001\014\001\015\001\003\000\016\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\055\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\017\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\018\001\019\001\
    \020\001\238\000\022\001\023\001\024\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\025\001\
    \026\001\027\001\238\000\003\000\029\001\056\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\030\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\031\001\032\001\033\001\
    \034\001\035\001\036\001\037\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\038\001\039\001\
    \040\001\238\000\003\000\042\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\057\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\043\001\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\044\001\045\001\046\001\238\000\
    \060\001\097\001\098\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\099\001\100\001\105\001\
    \106\001\003\000\107\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\058\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\108\001\114\001\
    \103\001\102\001\201\000\118\001\103\001\201\000\201\000\119\001\
    \092\001\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\201\000\063\001\111\001\120\001\103\001\
    \121\001\201\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\113\001\000\000\103\001\
    \003\000\000\000\059\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\116\001\
    \000\000\220\000\000\000\116\001\220\000\220\000\104\001\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\220\000\000\000\124\001\000\000\116\001\000\000\
    \220\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\255\255\000\000\000\000\116\001\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \050\001\000\000\000\000\050\001\050\001\117\001\059\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\050\001\000\000\000\000\000\000\000\000\000\000\050\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\070\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\112\001\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \065\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\255\255\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \063\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\074\001\074\001\074\001\074\001\074\001\074\001\074\001\
    \074\001\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\064\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\085\001\
    \085\001\085\001\085\001\085\001\085\001\085\001\085\001\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\066\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\089\001\000\000\000\000\089\001\089\001\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\089\001\000\000\000\000\000\000\000\000\
    \000\000\089\001\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \067\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\068\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \069\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\071\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\072\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\073\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\074\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\075\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\076\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\077\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\068\001\079\001\000\000\000\000\078\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\078\001\
    \000\000\000\000\000\000\067\001\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\079\000\096\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\096\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\095\001\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\115\000\
    \096\001\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\116\000\116\000\077\001\000\000\000\000\000\000\066\001\
    \000\000\000\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \114\000\000\000\114\000\000\000\000\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\107\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\
    \107\000\085\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \000\000\000\000\000\000\000\000\106\000\105\000\255\255\085\000\
    \000\000\255\255\085\000\105\000\085\000\000\000\000\000\085\000\
    \084\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\027\000\027\000\085\000\000\000\085\000\084\000\052\001\
    \100\000\084\000\000\000\084\000\106\000\105\000\084\000\000\000\
    \097\000\000\000\096\000\105\000\054\001\093\000\100\000\000\000\
    \084\000\000\000\084\000\095\000\084\000\085\000\092\000\101\000\
    \128\000\000\000\138\000\100\000\000\000\000\000\084\000\000\000\
    \100\000\084\000\095\000\084\000\207\000\091\000\084\000\000\000\
    \098\000\000\000\091\000\255\255\084\000\094\000\100\000\129\000\
    \000\000\139\000\084\000\095\000\084\000\000\000\092\000\000\000\
    \128\000\096\000\138\000\208\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\095\000\000\000\207\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\084\000\000\000\000\000\112\000\
    \000\000\112\000\000\000\100\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\111\000\111\000\107\000\107\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\108\000\105\000\000\000\110\000\000\000\
    \110\000\000\000\105\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\000\000\
    \000\000\000\000\000\000\108\000\105\000\000\000\000\000\000\000\
    \000\000\000\000\105\000\105\000\000\000\000\000\133\000\133\000\
    \000\000\105\000\111\000\111\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\113\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\000\000\000\000\
    \105\000\000\000\136\000\105\000\000\000\000\000\105\000\000\000\
    \000\000\105\000\105\000\135\000\127\000\000\000\127\000\000\000\
    \105\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\000\000\000\000\000\000\000\000\134\000\
    \105\000\000\000\137\000\000\000\000\000\255\255\105\000\000\000\
    \000\000\000\000\105\000\135\000\000\000\000\000\000\000\000\000\
    \105\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \000\000\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\000\000\000\000\000\000\000\000\122\000\000\000\116\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\000\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\000\000\000\000\
    \000\000\071\001\000\000\119\000\000\000\000\000\000\000\121\000\
    \125\000\000\000\125\000\000\000\118\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\000\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\000\000\000\000\
    \117\000\071\001\000\000\120\000\000\000\000\000\089\001\121\000\
    \000\000\089\001\089\001\000\000\118\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\089\001\
    \000\000\000\000\000\000\000\000\000\000\089\001\122\000\122\000\
    \122\000\122\000\122\000\122\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\123\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\105\000\000\000\000\000\000\000\
    \000\000\000\000\105\000\000\000\000\000\123\000\000\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\000\000\000\000\000\000\000\000\050\001\051\001\000\000\
    \050\001\050\001\000\000\000\000\105\000\105\000\000\000\000\000\
    \000\000\000\000\105\000\105\000\000\000\000\000\050\001\003\000\
    \000\000\000\000\000\000\000\000\050\001\000\000\049\001\000\000\
    \088\001\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\105\000\000\000\000\000\
    \000\000\255\255\000\000\105\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\173\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\255\255\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\174\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\255\255\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \175\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\176\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\193\000\194\000\000\000\193\000\193\000\201\000\
    \202\000\000\000\201\000\201\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\193\000\000\000\000\000\000\000\000\000\
    \201\000\193\000\000\000\189\000\000\000\000\000\201\000\000\000\
    \000\000\192\000\191\000\000\000\000\000\000\000\000\000\000\000\
    \190\000\000\000\255\255\000\000\000\000\255\255\200\000\200\000\
    \200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\
    \200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\
    \200\000\200\000\220\000\052\001\000\000\220\000\220\000\255\255\
    \000\000\255\255\255\255\000\000\255\255\000\000\000\000\000\000\
    \054\001\000\000\000\000\220\000\205\000\000\000\000\000\000\000\
    \000\000\220\000\000\000\000\000\000\000\204\000\000\000\000\000\
    \052\001\000\000\056\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\199\000\000\000\000\000\055\001\198\000\055\001\
    \000\000\203\000\000\000\000\000\206\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\204\000\000\000\000\000\
    \230\000\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\000\000\000\000\223\000\000\000\000\000\
    \000\000\000\000\072\001\072\001\072\001\072\001\072\001\072\001\
    \000\000\000\000\226\000\000\000\000\000\000\000\000\000\000\000\
    \222\000\000\000\229\000\000\000\000\000\221\000\000\000\000\000\
    \227\000\000\000\000\000\224\000\228\000\000\000\225\000\000\000\
    \231\000\000\000\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\072\001\072\001\072\001\072\001\072\001\072\001\081\001\
    \081\001\081\001\081\001\081\001\081\001\081\001\081\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \000\000\072\001\072\001\072\001\072\001\072\001\072\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\000\000\000\000\082\001\
    \000\000\255\255\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\083\001\
    \083\001\083\001\083\001\083\001\083\001\000\000\000\000\082\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\083\001\083\001\083\001\083\001\083\001\
    \083\001\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\083\001\083\001\083\001\083\001\083\001\
    \083\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\035\000\000\000\000\000\035\000\049\000\
    \255\255\255\255\049\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\149\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \014\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \017\000\017\000\018\000\019\000\020\000\022\000\033\000\025\000\
    \086\000\033\000\033\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\082\000\082\000\018\000\
    \021\000\018\000\018\000\020\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\043\000\040\000\
    \043\000\091\000\003\000\025\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\023\000\021\000\
    \024\000\038\000\046\000\099\000\046\000\104\000\038\000\117\000\
    \130\000\131\000\132\000\134\000\004\000\095\000\023\000\023\000\
    \140\000\141\000\024\000\142\000\038\000\040\000\143\000\146\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\147\000\148\000\095\000\129\000\118\000\165\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\118\000\095\000\129\000\118\000\004\000\
    \168\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\026\000\135\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\027\000\
    \177\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\037\000\119\000\182\000\037\000\037\000\
    \190\000\135\000\198\000\199\000\135\000\119\000\027\000\203\000\
    \136\000\209\000\087\000\210\000\037\000\027\000\041\000\139\000\
    \098\000\136\000\037\000\184\000\088\000\028\000\027\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\087\000\097\000\096\000\211\000\119\000\027\000\098\000\
    \087\000\028\000\027\000\088\000\028\000\027\000\041\000\139\000\
    \179\000\136\000\088\000\028\000\205\000\212\000\027\000\213\000\
    \100\000\097\000\096\000\088\000\028\000\205\000\208\000\028\000\
    \097\000\087\000\215\000\216\000\041\000\087\000\217\000\098\000\
    \087\000\028\000\096\000\088\000\028\000\218\000\098\000\100\000\
    \028\000\221\000\030\000\028\000\030\000\040\000\222\000\030\000\
    \223\000\097\000\096\000\088\000\028\000\205\000\208\000\028\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\096\000\224\000\225\000\226\000\204\000\100\000\
    \227\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\204\000\228\000\229\000\204\000\030\000\
    \230\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\032\000\110\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\232\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\231\000\233\000\234\000\235\000\231\000\236\000\237\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\239\000\240\000\241\000\242\000\032\000\243\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\053\000\244\000\041\000\245\000\179\000\193\000\
    \246\000\184\000\193\000\193\000\247\000\248\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \193\000\249\000\250\000\251\000\252\000\253\000\193\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\254\000\255\000\000\001\001\001\053\000\002\001\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\054\000\112\000\112\000\112\000\112\000\112\000\112\000\
    \112\000\112\000\112\000\112\000\003\001\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\004\001\
    \005\001\006\001\007\001\008\001\009\001\010\001\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \011\001\012\001\013\001\014\001\054\000\015\001\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \055\000\114\000\114\000\114\000\114\000\114\000\114\000\114\000\
    \114\000\114\000\114\000\016\001\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\017\001\018\001\
    \019\001\020\001\021\001\022\001\023\001\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\024\001\
    \025\001\026\001\027\001\055\000\028\001\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\056\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\029\001\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\030\001\031\001\032\001\
    \033\001\034\001\035\001\036\001\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\037\001\038\001\
    \039\001\040\001\056\000\041\001\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\057\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\042\001\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\043\001\044\001\045\001\046\001\
    \057\001\095\001\097\001\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\098\001\099\001\104\001\
    \105\001\057\000\106\001\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\058\000\107\001\113\001\
    \101\001\101\001\201\000\117\001\101\001\201\000\201\000\118\001\
    \090\001\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\201\000\061\001\109\001\119\001\101\001\
    \120\001\201\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\255\255\109\001\255\255\101\001\
    \058\000\255\255\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\059\000\255\255\255\255\115\001\
    \255\255\219\000\255\255\115\001\219\000\219\000\101\001\255\255\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\219\000\255\255\122\001\255\255\115\001\255\255\
    \219\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\090\001\255\255\255\255\115\001\059\000\
    \255\255\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\060\000\255\255\255\255\255\255\255\255\
    \050\001\255\255\255\255\050\001\050\001\115\001\057\001\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\050\001\255\255\255\255\255\255\255\255\255\255\050\001\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\255\255\255\255\255\255\255\255\060\000\255\255\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\061\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\061\001\109\001\255\255\255\255\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\255\255\255\255\255\255\255\255\061\000\255\255\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\062\000\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\122\001\255\255\255\255\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \255\255\255\255\255\255\255\255\062\000\255\255\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \063\000\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\255\255\255\255\255\255\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\255\255\
    \255\255\255\255\255\255\063\000\255\255\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\064\000\
    \081\001\081\001\081\001\081\001\081\001\081\001\081\001\081\001\
    \255\255\255\255\255\255\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\255\255\255\255\
    \255\255\255\255\064\000\255\255\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\065\000\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\255\255\
    \255\255\255\255\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\255\255\255\255\255\255\
    \255\255\065\000\255\255\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\066\000\255\255\255\255\
    \255\255\255\255\089\001\255\255\255\255\089\001\089\001\255\255\
    \255\255\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\089\001\255\255\255\255\255\255\255\255\
    \255\255\089\001\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\255\255\255\255\255\255\255\255\
    \066\000\255\255\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\067\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\255\255\255\255\255\255\255\255\067\000\
    \255\255\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\068\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\255\255\255\255\255\255\255\255\068\000\255\255\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\069\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\255\255\255\255\255\255\255\255\069\000\255\255\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\070\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \255\255\255\255\255\255\255\255\070\000\255\255\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \071\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\255\255\
    \255\255\255\255\255\255\071\000\255\255\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\072\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\255\255\255\255\
    \255\255\255\255\072\000\255\255\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\073\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\255\255\255\255\255\255\
    \255\255\073\000\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\074\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\255\255\255\255\255\255\255\255\
    \074\000\255\255\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\075\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\255\255\255\255\255\255\255\255\075\000\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\076\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\255\255\255\255\255\255\255\255\076\000\255\255\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\077\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\064\001\075\001\255\255\255\255\077\000\255\255\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\078\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\075\001\
    \255\255\255\255\255\255\064\001\255\255\255\255\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \255\255\255\255\255\255\255\255\078\000\093\001\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \079\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\093\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\255\255\
    \255\255\255\255\093\001\079\000\255\255\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\083\000\
    \093\001\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\075\001\255\255\255\255\255\255\064\001\
    \255\255\255\255\083\000\083\000\083\000\083\000\083\000\083\000\
    \084\000\255\255\084\000\255\255\255\255\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\089\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \255\255\255\255\255\255\255\255\085\000\085\000\049\001\101\000\
    \255\255\049\001\092\000\085\000\093\000\255\255\255\255\090\000\
    \089\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\094\000\255\255\102\000\101\000\049\001\
    \089\000\092\000\255\255\093\000\085\000\085\000\090\000\255\255\
    \092\000\255\255\093\000\085\000\049\001\090\000\101\000\255\255\
    \089\000\255\255\094\000\093\000\102\000\103\000\090\000\089\000\
    \120\000\255\255\137\000\102\000\255\255\255\255\101\000\255\255\
    \089\000\092\000\094\000\093\000\206\000\092\000\090\000\255\255\
    \092\000\255\255\090\000\093\001\103\000\090\000\101\000\120\000\
    \255\255\137\000\094\000\093\000\102\000\255\255\090\000\255\255\
    \120\000\094\000\137\000\206\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\094\000\255\255\206\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\103\000\255\255\255\255\106\000\
    \255\255\106\000\255\255\103\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\107\000\107\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\107\000\107\000\255\255\108\000\255\255\
    \108\000\255\255\107\000\108\000\108\000\108\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\255\255\
    \255\255\255\255\255\255\107\000\107\000\255\255\255\255\255\255\
    \255\255\255\255\107\000\109\000\255\255\255\255\133\000\133\000\
    \255\255\109\000\111\000\111\000\111\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\113\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\255\255\255\255\
    \111\000\255\255\133\000\109\000\255\255\255\255\111\000\255\255\
    \255\255\109\000\113\000\133\000\121\000\255\255\121\000\255\255\
    \113\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\255\255\255\255\255\255\255\255\133\000\
    \111\000\255\255\133\000\255\255\255\255\049\001\111\000\255\255\
    \255\255\255\255\113\000\133\000\255\255\255\255\255\255\255\255\
    \113\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\115\000\115\000\115\000\115\000\115\000\115\000\
    \255\255\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\255\255\255\255\255\255\255\255\116\000\255\255\116\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\255\255\115\000\115\000\115\000\115\000\115\000\115\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\255\255\255\255\
    \255\255\067\001\255\255\116\000\255\255\255\255\255\255\116\000\
    \123\000\255\255\123\000\255\255\116\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\255\255\
    \116\000\116\000\116\000\116\000\116\000\116\000\255\255\255\255\
    \116\000\067\001\255\255\116\000\255\255\255\255\086\001\116\000\
    \255\255\086\001\086\001\255\255\116\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\086\001\
    \255\255\255\255\255\255\255\255\255\255\086\001\122\000\122\000\
    \122\000\122\000\122\000\122\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\122\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\122\000\122\000\
    \122\000\122\000\122\000\122\000\124\000\255\255\255\255\255\255\
    \255\255\255\255\124\000\255\255\255\255\122\000\255\255\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\255\255\255\255\255\255\255\255\047\001\047\001\255\255\
    \047\001\047\001\255\255\255\255\124\000\126\000\255\255\255\255\
    \255\255\255\255\124\000\126\000\255\255\255\255\047\001\172\000\
    \255\255\255\255\255\255\255\255\047\001\255\255\047\001\255\255\
    \086\001\255\255\255\255\172\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\126\000\255\255\255\255\
    \255\255\067\001\255\255\126\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\255\255\255\255\
    \255\255\255\255\172\000\255\255\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\
    \172\000\172\000\172\000\172\000\172\000\172\000\173\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\086\001\255\255\255\255\
    \255\255\173\000\255\255\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\173\000\174\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\174\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\047\001\255\255\255\255\255\255\
    \174\000\255\255\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\175\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\255\255\255\255\255\255\255\255\175\000\
    \255\255\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\
    \175\000\175\000\175\000\176\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\255\255\255\255\255\255\255\255\176\000\255\255\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\187\000\187\000\255\255\187\000\187\000\196\000\
    \196\000\255\255\196\000\196\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\187\000\255\255\255\255\255\255\255\255\
    \196\000\187\000\255\255\187\000\255\255\255\255\196\000\255\255\
    \255\255\187\000\187\000\255\255\255\255\255\255\255\255\255\255\
    \187\000\255\255\053\001\255\255\255\255\053\001\196\000\196\000\
    \196\000\196\000\196\000\196\000\196\000\196\000\196\000\196\000\
    \200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\
    \200\000\200\000\220\000\053\001\255\255\220\000\220\000\054\001\
    \255\255\055\001\054\001\255\255\055\001\255\255\255\255\255\255\
    \053\001\255\255\255\255\220\000\200\000\255\255\255\255\255\255\
    \255\255\220\000\255\255\255\255\255\255\200\000\255\255\255\255\
    \054\001\255\255\055\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\196\000\255\255\255\255\054\001\196\000\055\001\
    \255\255\200\000\255\255\255\255\200\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\200\000\255\255\255\255\
    \220\000\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\255\255\255\255\220\000\255\255\255\255\
    \255\255\255\255\071\001\071\001\071\001\071\001\071\001\071\001\
    \255\255\255\255\220\000\255\255\255\255\255\255\255\255\255\255\
    \220\000\255\255\220\000\255\255\255\255\220\000\255\255\255\255\
    \220\000\255\255\255\255\220\000\220\000\255\255\220\000\255\255\
    \220\000\255\255\071\001\071\001\071\001\071\001\071\001\071\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\072\001\072\001\072\001\072\001\072\001\072\001\078\001\
    \078\001\078\001\078\001\078\001\078\001\078\001\078\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \255\255\072\001\072\001\072\001\072\001\072\001\072\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\255\255\255\255\078\001\
    \255\255\187\000\255\255\255\255\255\255\255\255\196\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\082\001\
    \082\001\082\001\082\001\082\001\082\001\255\255\255\255\078\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\053\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\083\001\083\001\083\001\083\001\083\001\
    \083\001\255\255\255\255\255\255\255\255\255\255\054\001\255\255\
    \055\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\083\001\083\001\083\001\083\001\083\001\
    \083\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\078\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\001\000\001\000\001\000\001\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\001\000\000\000\000\000\000\000\
    \001\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\219\000\255\255\220\000\219\000\219\000\220\000\220\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\219\000\255\255\220\000\255\255\255\255\255\255\
    \219\000\255\255\220\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\001\255\255\000\001\255";
}

let rec initial lexbuf =
    __ocaml_lex_initial_rec lexbuf 0
and __ocaml_lex_initial_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 470 "src/kernel_internals/parsing/clexer.mll"
      (
        let s = Lexing.lexeme lexbuf in
        let first_string =
          if String.length s > 2 then
            String.sub s 2 (String.length s - 2)
          else ""
        in
	do_lex_comment ~first_string comment lexbuf ;
        initial lexbuf
      )
# 1855 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 481 "src/kernel_internals/parsing/clexer.mll"
                        c
# 1861 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 482 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
      try
        save_current_pos ();
	Buffer.clear buf;
	annot_first_token lexbuf
      with Parsing.Parse_error when Kernel.ContinueOnAnnotError.get () ->
        let source = Lexing.lexeme_start_p lexbuf in
        Kernel.debug ~source "skipping annotation";
	initial lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) comment lexbuf ;
          initial lexbuf
	end
    )
# 1879 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 499 "src/kernel_internals/parsing/clexer.mll"
    ( 
      let s = Lexing.lexeme lexbuf in
      let first_string =
        if String.length s > 2 then
          String.sub s 2 (String.length s - 2)
        else ""
      in
      do_lex_comment ~first_string onelinecomment lexbuf ;
      E.newline();
      if is_oneline_ghost () then begin
        exit_oneline_ghost ();
        RGHOST
      end else begin
        initial lexbuf
      end
    )
# 1899 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 516 "src/kernel_internals/parsing/clexer.mll"
                    c
# 1905 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 517 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
      try
        save_current_pos ();
	Buffer.clear buf;
	annot_one_line lexbuf
      with Parsing.Parse_error when Kernel.ContinueOnAnnotError.get () ->
        let source = Lexing.lexeme_start_p lexbuf in
        Kernel.debug ~source "skipping annotation";
        initial lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) onelinecomment lexbuf;
	  E.newline();
	  if is_oneline_ghost () then
	    begin
              exit_oneline_ghost ();
              RGHOST
	    end
	  else
	    begin
              initial lexbuf
	    end
	end
    )
# 1932 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 541 "src/kernel_internals/parsing/clexer.mll"
           (initial lexbuf)
# 1937 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 542 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          if !pragmaLine then
                                            begin
                                              pragmaLine := false;
                                              PRAGMA_EOL
                                            end
                                          else if is_oneline_ghost () then
                                            begin
                                              exit_oneline_ghost ();
                                              RGHOST
                                            end
                                          else begin
                                            initial lexbuf
                                          end )
# 1955 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 556 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          initial lexbuf
                                        )
# 1962 "src/kernel_internals/parsing/clexer.ml"

  | 7 ->
# 559 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 1967 "src/kernel_internals/parsing/clexer.ml"

  | 8 ->
# 560 "src/kernel_internals/parsing/clexer.mll"
          ( hash lexbuf)
# 1972 "src/kernel_internals/parsing/clexer.ml"

  | 9 ->
# 561 "src/kernel_internals/parsing/clexer.mll"
                                   ( PRAGMA (currentLoc ()) )
# 1977 "src/kernel_internals/parsing/clexer.ml"

  | 10 ->
# 562 "src/kernel_internals/parsing/clexer.mll"
          ( CST_CHAR (chr lexbuf, currentLoc ()))
# 1982 "src/kernel_internals/parsing/clexer.ml"

  | 11 ->
# 563 "src/kernel_internals/parsing/clexer.mll"
          ( CST_WCHAR (chr lexbuf, currentLoc ()) )
# 1987 "src/kernel_internals/parsing/clexer.ml"

  | 12 ->
# 564 "src/kernel_internals/parsing/clexer.mll"
         (  
(* matth: BUG:  this could be either a regular string or a wide string.
 *  e.g. if it's the "world" in
 *     L"Hello, " "world"
 *  then it should be treated as wide even though there's no L immediately
 *  preceding it.  See test/small1/wchar5.c for a failure case. *)
                                          CST_STRING (str lexbuf, currentLoc()))
# 1998 "src/kernel_internals/parsing/clexer.ml"

  | 13 ->
# 571 "src/kernel_internals/parsing/clexer.mll"
           ( (* weimer: wchar_t string literal *)
                                          CST_WSTRING(str lexbuf, currentLoc()))
# 2004 "src/kernel_internals/parsing/clexer.ml"

  | 14 ->
# 573 "src/kernel_internals/parsing/clexer.mll"
             (CST_FLOAT (Lexing.lexeme lexbuf, currentLoc ()))
# 2009 "src/kernel_internals/parsing/clexer.ml"

  | 15 ->
# 574 "src/kernel_internals/parsing/clexer.mll"
                           ( (* GCC Extension for binary numbers *) 
                                          CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2015 "src/kernel_internals/parsing/clexer.ml"

  | 16 ->
# 576 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2020 "src/kernel_internals/parsing/clexer.ml"

  | 17 ->
# 577 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2025 "src/kernel_internals/parsing/clexer.ml"

  | 18 ->
# 578 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2030 "src/kernel_internals/parsing/clexer.ml"

  | 19 ->
# 579 "src/kernel_internals/parsing/clexer.mll"
             (EOF)
# 2035 "src/kernel_internals/parsing/clexer.ml"

  | 20 ->
# 580 "src/kernel_internals/parsing/clexer.mll"
           (ELLIPSIS)
# 2040 "src/kernel_internals/parsing/clexer.ml"

  | 21 ->
# 581 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_EQ)
# 2045 "src/kernel_internals/parsing/clexer.ml"

  | 22 ->
# 582 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_EQ)
# 2050 "src/kernel_internals/parsing/clexer.ml"

  | 23 ->
# 583 "src/kernel_internals/parsing/clexer.mll"
          (STAR_EQ)
# 2055 "src/kernel_internals/parsing/clexer.ml"

  | 24 ->
# 584 "src/kernel_internals/parsing/clexer.mll"
          (SLASH_EQ)
# 2060 "src/kernel_internals/parsing/clexer.ml"

  | 25 ->
# 585 "src/kernel_internals/parsing/clexer.mll"
          (PERCENT_EQ)
# 2065 "src/kernel_internals/parsing/clexer.ml"

  | 26 ->
# 586 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_EQ)
# 2070 "src/kernel_internals/parsing/clexer.ml"

  | 27 ->
# 587 "src/kernel_internals/parsing/clexer.mll"
          (AND_EQ)
# 2075 "src/kernel_internals/parsing/clexer.ml"

  | 28 ->
# 588 "src/kernel_internals/parsing/clexer.mll"
          (CIRC_EQ)
# 2080 "src/kernel_internals/parsing/clexer.ml"

  | 29 ->
# 589 "src/kernel_internals/parsing/clexer.mll"
           (INF_INF_EQ)
# 2085 "src/kernel_internals/parsing/clexer.ml"

  | 30 ->
# 590 "src/kernel_internals/parsing/clexer.mll"
           (SUP_SUP_EQ)
# 2090 "src/kernel_internals/parsing/clexer.ml"

  | 31 ->
# 591 "src/kernel_internals/parsing/clexer.mll"
          (INF_INF)
# 2095 "src/kernel_internals/parsing/clexer.ml"

  | 32 ->
# 592 "src/kernel_internals/parsing/clexer.mll"
          (SUP_SUP)
# 2100 "src/kernel_internals/parsing/clexer.ml"

  | 33 ->
# 593 "src/kernel_internals/parsing/clexer.mll"
           (EQ_EQ)
# 2105 "src/kernel_internals/parsing/clexer.ml"

  | 34 ->
# 594 "src/kernel_internals/parsing/clexer.mll"
           (EXCLAM_EQ)
# 2110 "src/kernel_internals/parsing/clexer.ml"

  | 35 ->
# 595 "src/kernel_internals/parsing/clexer.mll"
          (INF_EQ)
# 2115 "src/kernel_internals/parsing/clexer.ml"

  | 36 ->
# 596 "src/kernel_internals/parsing/clexer.mll"
          (SUP_EQ)
# 2120 "src/kernel_internals/parsing/clexer.ml"

  | 37 ->
# 597 "src/kernel_internals/parsing/clexer.mll"
         (EQ)
# 2125 "src/kernel_internals/parsing/clexer.ml"

  | 38 ->
# 598 "src/kernel_internals/parsing/clexer.mll"
         (INF)
# 2130 "src/kernel_internals/parsing/clexer.ml"

  | 39 ->
# 599 "src/kernel_internals/parsing/clexer.mll"
         (SUP)
# 2135 "src/kernel_internals/parsing/clexer.ml"

  | 40 ->
# 600 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_PLUS (currentLoc ()))
# 2140 "src/kernel_internals/parsing/clexer.ml"

  | 41 ->
# 601 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_MINUS (currentLoc ()))
# 2145 "src/kernel_internals/parsing/clexer.ml"

  | 42 ->
# 602 "src/kernel_internals/parsing/clexer.mll"
          (ARROW)
# 2150 "src/kernel_internals/parsing/clexer.ml"

  | 43 ->
# 603 "src/kernel_internals/parsing/clexer.mll"
         (PLUS (currentLoc ()))
# 2155 "src/kernel_internals/parsing/clexer.ml"

  | 44 ->
# 604 "src/kernel_internals/parsing/clexer.mll"
         (MINUS (currentLoc ()))
# 2160 "src/kernel_internals/parsing/clexer.ml"

  | 45 ->
# 606 "src/kernel_internals/parsing/clexer.mll"
                    ( if is_ghost_code () then might_end_ghost lexbuf
                      else
                        STAR (currentLoc ()))
# 2167 "src/kernel_internals/parsing/clexer.ml"

  | 46 ->
# 609 "src/kernel_internals/parsing/clexer.mll"
         (SLASH)
# 2172 "src/kernel_internals/parsing/clexer.ml"

  | 47 ->
# 610 "src/kernel_internals/parsing/clexer.mll"
         (PERCENT)
# 2177 "src/kernel_internals/parsing/clexer.ml"

  | 48 ->
# 611 "src/kernel_internals/parsing/clexer.mll"
         (EXCLAM (currentLoc ()))
# 2182 "src/kernel_internals/parsing/clexer.ml"

  | 49 ->
# 612 "src/kernel_internals/parsing/clexer.mll"
          (AND_AND (currentLoc ()))
# 2187 "src/kernel_internals/parsing/clexer.ml"

  | 50 ->
# 613 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_PIPE)
# 2192 "src/kernel_internals/parsing/clexer.ml"

  | 51 ->
# 614 "src/kernel_internals/parsing/clexer.mll"
         (AND (currentLoc ()))
# 2197 "src/kernel_internals/parsing/clexer.ml"

  | 52 ->
# 615 "src/kernel_internals/parsing/clexer.mll"
         (PIPE)
# 2202 "src/kernel_internals/parsing/clexer.ml"

  | 53 ->
# 616 "src/kernel_internals/parsing/clexer.mll"
         (CIRC)
# 2207 "src/kernel_internals/parsing/clexer.ml"

  | 54 ->
# 617 "src/kernel_internals/parsing/clexer.mll"
         (QUEST)
# 2212 "src/kernel_internals/parsing/clexer.ml"

  | 55 ->
# 619 "src/kernel_internals/parsing/clexer.mll"
  ( if Cabshelper.is_attr_test () then begin
        Cabshelper.pop_attr_test (); COLON2
      end else COLON
  )
# 2220 "src/kernel_internals/parsing/clexer.ml"

  | 56 ->
# 623 "src/kernel_internals/parsing/clexer.mll"
                (TILDE (currentLoc ()))
# 2225 "src/kernel_internals/parsing/clexer.ml"

  | 57 ->
# 625 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LBRACE (currentLoc ())))
# 2230 "src/kernel_internals/parsing/clexer.ml"

  | 58 ->
# 626 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (RBRACE (currentLoc ())))
# 2235 "src/kernel_internals/parsing/clexer.ml"

  | 59 ->
# 627 "src/kernel_internals/parsing/clexer.mll"
                (dbgToken (LBRACE (currentLoc ())))
# 2240 "src/kernel_internals/parsing/clexer.ml"

  | 60 ->
# 628 "src/kernel_internals/parsing/clexer.mll"
                (dbgToken (RBRACE (currentLoc ())))
# 2245 "src/kernel_internals/parsing/clexer.ml"

  | 61 ->
# 629 "src/kernel_internals/parsing/clexer.mll"
          (LBRACKET)
# 2250 "src/kernel_internals/parsing/clexer.ml"

  | 62 ->
# 630 "src/kernel_internals/parsing/clexer.mll"
          (RBRACKET)
# 2255 "src/kernel_internals/parsing/clexer.ml"

  | 63 ->
# 631 "src/kernel_internals/parsing/clexer.mll"
           (LBRACKET)
# 2260 "src/kernel_internals/parsing/clexer.ml"

  | 64 ->
# 632 "src/kernel_internals/parsing/clexer.mll"
           (RBRACKET)
# 2265 "src/kernel_internals/parsing/clexer.ml"

  | 65 ->
# 633 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LPAREN (currentLoc ())) )
# 2270 "src/kernel_internals/parsing/clexer.ml"

  | 66 ->
# 634 "src/kernel_internals/parsing/clexer.mll"
          (RPAREN)
# 2275 "src/kernel_internals/parsing/clexer.ml"

  | 67 ->
# 635 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (SEMICOLON (currentLoc ())) )
# 2280 "src/kernel_internals/parsing/clexer.ml"

  | 68 ->
# 636 "src/kernel_internals/parsing/clexer.mll"
          (COMMA)
# 2285 "src/kernel_internals/parsing/clexer.ml"

  | 69 ->
# 637 "src/kernel_internals/parsing/clexer.mll"
          (DOT)
# 2290 "src/kernel_internals/parsing/clexer.ml"

  | 70 ->
# 638 "src/kernel_internals/parsing/clexer.mll"
             (SIZEOF (currentLoc ()))
# 2295 "src/kernel_internals/parsing/clexer.ml"

  | 71 ->
# 639 "src/kernel_internals/parsing/clexer.mll"
                                        ( if !Cprint.msvcMode then
                                             MSASM (msasm lexbuf, currentLoc ())
                                          else (ASM (currentLoc ())) )
# 2302 "src/kernel_internals/parsing/clexer.ml"

  | 72 ->
# 644 "src/kernel_internals/parsing/clexer.mll"
                                        ( let _ = matchingpars 0 lexbuf in
                                          initial lexbuf
                                        )
# 2309 "src/kernel_internals/parsing/clexer.ml"

  | 73 ->
# 650 "src/kernel_internals/parsing/clexer.mll"
                                        (initial lexbuf )
# 2314 "src/kernel_internals/parsing/clexer.ml"

  | 74 ->
# 651 "src/kernel_internals/parsing/clexer.mll"
           (scan_ident (Lexing.lexeme lexbuf))
# 2319 "src/kernel_internals/parsing/clexer.ml"

  | 75 ->
# 653 "src/kernel_internals/parsing/clexer.mll"
  ( if is_oneline_ghost() then begin
      exit_oneline_ghost (); RGHOST
    end
    else EOF
  )
# 2328 "src/kernel_internals/parsing/clexer.ml"

  | 76 ->
# 658 "src/kernel_internals/parsing/clexer.mll"
       (E.parse_error "Invalid symbol")
# 2333 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_initial_rec lexbuf __ocaml_lex_state

and might_end_ghost lexbuf =
    __ocaml_lex_might_end_ghost_rec lexbuf 177
and __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 661 "src/kernel_internals/parsing/clexer.mll"
        ( exit_ghost_code(); RGHOST )
# 2345 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 662 "src/kernel_internals/parsing/clexer.mll"
       ( STAR (currentLoc()) )
# 2350 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state

and comment buffer lexbuf =
    __ocaml_lex_comment_rec buffer lexbuf 179
and __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 665 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2362 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 666 "src/kernel_internals/parsing/clexer.mll"
                ( E.parse_error "Unterminated C comment" )
# 2367 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 667 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment comment buffer lexbuf )
# 2372 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state

and onelinecomment buffer lexbuf =
    __ocaml_lex_onelinecomment_rec buffer lexbuf 184
and __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 670 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2384 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 671 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment onelinecomment buffer lexbuf )
# 2389 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state

and matchingpars parsopen lexbuf =
    __ocaml_lex_matchingpars_rec parsopen lexbuf 187
and __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 674 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); matchingpars parsopen lexbuf )
# 2401 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 675 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2406 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 676 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars (parsopen + 1) lexbuf )
# 2411 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 677 "src/kernel_internals/parsing/clexer.mll"
                ( if parsopen > 1 then
                    matchingpars (parsopen - 1) lexbuf
                )
# 2418 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 680 "src/kernel_internals/parsing/clexer.mll"
         ( do_lex_comment comment lexbuf ;
                  matchingpars parsopen lexbuf )
# 2424 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 682 "src/kernel_internals/parsing/clexer.mll"
        ( let _ = str lexbuf in
                  matchingpars parsopen lexbuf )
# 2430 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 684 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2435 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state

and hash lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1) ;   __ocaml_lex_hash_rec lexbuf 196
and __ocaml_lex_hash_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 688 "src/kernel_internals/parsing/clexer.mll"
        ( E.newline (); initial lexbuf)
# 2447 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 689 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 2452 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 690 "src/kernel_internals/parsing/clexer.mll"
         ( (* We are seeing a line number. This is the number for the
                   * next line *)
                 let s = Lexing.lexeme lexbuf in
                 let lineno = try
                   int_of_string s
                 with Failure _ ->
                   (* the int is too big. *)
                   Kernel.warning "Bad line number in preprocessed file: %s" s;
                   (-1)
                 in
                 E.setCurrentLine (lineno - 1);
                  (* A file name may follow *)
		  file lexbuf )
# 2469 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 703 "src/kernel_internals/parsing/clexer.mll"
                ( hash lexbuf )
# 2474 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
let
# 706 "src/kernel_internals/parsing/clexer.mll"
                                     pragmaName
# 2480 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 707 "src/kernel_internals/parsing/clexer.mll"
                ( let here = currentLoc () in
                  PRAGMA_LINE (pragmaName ^ pragma lexbuf, here)
                )
# 2486 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 710 "src/kernel_internals/parsing/clexer.mll"
                ( pragmaLine := true; PRAGMA (currentLoc ()) )
# 2491 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 711 "src/kernel_internals/parsing/clexer.mll"
            ( endline lexbuf)
# 2496 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_hash_rec lexbuf __ocaml_lex_state

and file lexbuf =
    __ocaml_lex_file_rec lexbuf 303
and __ocaml_lex_file_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 714 "src/kernel_internals/parsing/clexer.mll"
                      (E.newline (); initial lexbuf)
# 2508 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 715 "src/kernel_internals/parsing/clexer.mll"
          (file lexbuf)
# 2513 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 718 "src/kernel_internals/parsing/clexer.mll"
                                             (
                                 let n = Lexing.lexeme lexbuf in
				 let n1 = String.sub n 1 ((String.length n) - 4) in
                                 E.setCurrentWorkingDirectory n1;
                                 endline lexbuf )
# 2522 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 723 "src/kernel_internals/parsing/clexer.mll"
                                (  (* '"' *)
                                   let n = Lexing.lexeme lexbuf in
                                   let n1 = String.sub n 1
                                       ((String.length n) - 2) in
                                   E.setCurrentFile n1;
				 endline lexbuf)
# 2532 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 730 "src/kernel_internals/parsing/clexer.mll"
      (endline lexbuf)
# 2537 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_file_rec lexbuf __ocaml_lex_state

and endline lexbuf =
    __ocaml_lex_endline_rec lexbuf 313
and __ocaml_lex_endline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 733 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); initial lexbuf)
# 2549 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 734 "src/kernel_internals/parsing/clexer.mll"
                                ( EOF )
# 2554 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 735 "src/kernel_internals/parsing/clexer.mll"
      ( endline lexbuf)
# 2559 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_endline_rec lexbuf __ocaml_lex_state

and pragma lexbuf =
    __ocaml_lex_pragma_rec lexbuf 317
and __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 738 "src/kernel_internals/parsing/clexer.mll"
                        ( E.newline (); "" )
# 2571 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 739 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (pragma lexbuf) )
# 2577 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state

and str lexbuf =
    __ocaml_lex_str_rec lexbuf 320
and __ocaml_lex_str_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 743 "src/kernel_internals/parsing/clexer.mll"
                        ([])
# 2589 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 744 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape str lexbuf)
# 2594 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 745 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape str lexbuf)
# 2599 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 746 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape str lexbuf)
# 2604 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 747 "src/kernel_internals/parsing/clexer.mll"
                        (E.parse_error "unterminated string" )
# 2609 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 748 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped str lexbuf)
# 2614 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_str_rec lexbuf __ocaml_lex_state

and chr lexbuf =
    __ocaml_lex_chr_rec lexbuf 331
and __ocaml_lex_chr_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 751 "src/kernel_internals/parsing/clexer.mll"
              ([])
# 2626 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 752 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape chr lexbuf)
# 2631 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 753 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape chr lexbuf)
# 2636 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 754 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape chr lexbuf)
# 2641 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 755 "src/kernel_internals/parsing/clexer.mll"
                        ( E.parse_error "unterminated char" )
# 2646 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 756 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped chr lexbuf)
# 2651 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_chr_rec lexbuf __ocaml_lex_state

and msasm lexbuf =
    __ocaml_lex_msasm_rec lexbuf 342
and __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 759 "src/kernel_internals/parsing/clexer.mll"
                        ( msasm lexbuf )
# 2663 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 760 "src/kernel_internals/parsing/clexer.mll"
                        ( msasminbrace lexbuf )
# 2668 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 761 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasmnobrace lexbuf) )
# 2674 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state

and msasminbrace lexbuf =
    __ocaml_lex_msasminbrace_rec lexbuf 346
and __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 765 "src/kernel_internals/parsing/clexer.mll"
                        ( "" )
# 2686 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 766 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasminbrace lexbuf) )
# 2692 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state

and msasmnobrace lexbuf =
    __ocaml_lex_msasmnobrace_rec lexbuf 349
and __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 769 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 1;
                          "" )
# 2706 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 772 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 5;
                          "" )
# 2713 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 775 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in

                          cur ^ (msasmnobrace lexbuf) )
# 2720 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state

and annot_first_token lexbuf =
    __ocaml_lex_annot_first_token_rec lexbuf 357
and __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 780 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      Buffer.clear buf;
      enter_ghost_code ();
      LGHOST
    )
# 2737 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 786 "src/kernel_internals/parsing/clexer.mll"
                         c
# 2743 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 786 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_first_token lexbuf )
# 2747 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 787 "src/kernel_internals/parsing/clexer.mll"
         ( E.newline(); Buffer.add_char buf '\n'; annot_first_token lexbuf )
# 2752 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 788 "src/kernel_internals/parsing/clexer.mll"
       ( annot_token lexbuf )
# 2757 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state

and annot_token lexbuf =
    __ocaml_lex_annot_token_rec lexbuf 365
and __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 790 "src/kernel_internals/parsing/clexer.mll"
         ( let s = Buffer.contents buf in
           make_annot ~one_line:false lexbuf s )
# 2770 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 792 "src/kernel_internals/parsing/clexer.mll"
         ( E.parse_error "Unterminated annotation" )
# 2775 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 793 "src/kernel_internals/parsing/clexer.mll"
         (E.newline(); Buffer.add_char buf '\n'; annot_token lexbuf )
# 2780 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 794 "src/kernel_internals/parsing/clexer.mll"
         c
# 2786 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 794 "src/kernel_internals/parsing/clexer.mll"
           ( Buffer.add_char buf c; annot_token lexbuf )
# 2790 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state

and annot_one_line lexbuf =
    __ocaml_lex_annot_one_line_rec lexbuf 371
and __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 797 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      enter_oneline_ghost (); LGHOST
    )
# 2805 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 801 "src/kernel_internals/parsing/clexer.mll"
                         c
# 2811 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 801 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_one_line lexbuf )
# 2815 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 802 "src/kernel_internals/parsing/clexer.mll"
       ( annot_one_line_logic lexbuf )
# 2820 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state

and annot_one_line_logic lexbuf =
    __ocaml_lex_annot_one_line_logic_rec lexbuf 378
and __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 804 "src/kernel_internals/parsing/clexer.mll"
         ( make_annot ~one_line:true lexbuf (Buffer.contents buf) )
# 2832 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 805 "src/kernel_internals/parsing/clexer.mll"
         c
# 2838 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 805 "src/kernel_internals/parsing/clexer.mll"
           ( Buffer.add_char buf c; annot_one_line_logic lexbuf )
# 2842 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state

;;

# 807 "src/kernel_internals/parsing/clexer.mll"
 

  (* Catch the exceptions raised by the lexer itself *)
  let initial lexbuf =
    try initial lexbuf
    with Failure _ -> raise Parsing.Parse_error 


# 2858 "src/kernel_internals/parsing/clexer.ml"
